import java.awt.*;
import javax.swing.*;
import javax.swing.border.LineBorder;

/**
巡回セールスマン問題に対する局所探索法のプログラム例
http://web1.kcg.edu/~t_ibaraki/today/tsp.c
を Java に移植した。
*/

public class TSP extends GraphApplication {
    private int tourLength, tour[], optimal;
    protected JLabel label;
    private boolean initPressed = false;
    final static Color EDGE_COLOR2 = Color.cyan;
    private JLabel initLabel;

    public TSP(String title, String file, int optimal) {
        super(title, file);
        this.optimal = optimal;
        messagePanel.setLayout( new GridLayout( 4, 0 ) );
    }

    public void drawAllEdges( Graphics g ) {
        if (tour == null) return;
        for ( int i=0; i<n; i++ ) {
            drawEdge( tour[i % n], tour[(i+1) % n], g );
        }
    }

    private void draw( int i, int j ) {
        drawEdge( tour[i % n], tour[j % n] );
    }

    private void draw( int i ) {
        drawNodeCircle( tour[i % n] );
    }

    /** アルゴリズムの進度に応じて表示する内容をセットする */
    private void newLabel( String title, Color bg ) {
        messagePanel.add( label = new JLabel() );
        label.setHorizontalAlignment( JLabel.CENTER );
        label.setForeground( FG_COLOR );
        label.setFont(FONT);
        label.setBorder(new LineBorder(bg, 2, true));
        if (initLabel != null) label.setPreferredSize(initLabel.getSize());
        drawTourLength( title );
        if (initLabel == null) {
            initLabel = label;
            messagePanel.validate();  // 配置しなおすと線がずれるので１度だけ
        }
    }

    private void drawTourLength( String s ) {
        if (optimal == 0) {  label.setText(s);  return;  }
        if (tourLength == 0) {
            label.setText( "[" + s + "] tour length = *%*%*%*%* (+****.*%)" );
        } else {
            long rate = (long)( tourLength - optimal ) * 1000 / optimal;
            label.setText( "[" + s + "] tour length = " + tourLength
                           + " (+" + (rate/10.0) + "%)" );
        }
    }
    
    private void initTourLength() {
        tourLength = 0;
        for ( int i=0; i<n; i++ ) {
            tourLength += d[tour[i]][tour[(i + 1) % n]];
        }
    }

    private void initTourRandom() {
        tour = new int[n];
        for ( int i=0; i<n; i++ ) tour[i] = i;
        for ( int i=0; i<n; i++ ) {
            int j = (int)(Math.random() * n);
            int swap = tour[i];  tour[i] = tour[j];  tour[j] = swap;
        }
        initTourLength();
    }

    private void initNoTour() {
        tour = null;
        tourLength = 0;
    }

    private boolean twoOpt(final int a) {
        final int a1 = tour[a % n],  a2 = tour[(a + 1) % n];
        for ( int b=a+2; b<a+n-1; b++ ) {
            final int b1 = tour[b % n],  b2 = tour[(b + 1) % n];
            int diff = - (d[a1][a2] + d[b1][b2])
                       + (d[a1][b1] + d[a2][b2]);
            if ( diff < 0 ) {
                tourLength += diff;
                twoOptModify(a, b);
                return true;
            }
        }
        return false;
    }

    private void twoOptModify(int a, int b) {
        boolean drawn = false;
        if ( threadSuspended ) {
            drawn = true;
            draw(a); draw(a+1);
            draw(b); draw(b+1);
            if (!sleep()) { initPressed = true; return; }
            
            setColors(null, null, EDGE_COLOR2);
            draw(a, a+1);  draw(b, b+1);
            setColors(null, null, getBackground());
            if (!sleep()) { initPressed = true; return; }

            setColors(null, null, EDGE_COLOR2);
            draw(a, b);  draw(a+1, b+1);
            setColors(null, null, getBackground());
            if (!sleep()) { initPressed = true; return; }

            setColors(null, null, EDGE_COLOR2);
            draw(a, a+1);  draw(b, b+1);
            draw(a, b);  draw(a+1, b+1);
            setColors(null, null, getBackground());
        }
        drawTourLength("2-opt");
        draw(a, a+1);  draw(b, b+1);
        draw(a, b);  draw(a+1, b+1);
        if ( threadSuspended || drawn ) {
            if (!sleep()) { initPressed = true; return; }
            draw(a); draw(a+1);
            draw(b); draw(b+1);
        }
        for (int i = 0; i < (b - a) / 2; i++) {
            int swap = tour[(a + 1 + i) % n];
            tour[(a + 1 + i) % n] = tour[(b - i) % n]; 
            tour[(b - i) % n] = swap;
        }
        if ( threadSuspended || drawn ) repaint();
    }

    private boolean orOpt(final int a) {
        final int a1 = tour[a % n],  a2 = tour[(a + 1) % n];
        for (int b = a + 1; b <= a + 4; b++) {
            final int b1 = tour[b % n],  b2 = tour[(b + 1) % n];
            for (int c = b + 1; c < a + n - 1; c++) {
                final int c1 = tour[c % n],  c2 = tour[(c + 1) % n];
                final int before = d[a1][a2] + d[b1][b2] + d[c1][c2];
                int after =        d[a1][b2] + d[a2][c2] + d[b1][c1];
                if (after - before < 0) {
                    tourLength += after - before;
                    orOptModify(a, b, c, false);
                    return true;
                }
                if (b == a + 1) continue;
                after = d[a1][b2] + d[a2][c1] + d[b1][c2];
                if (after - before < 0) {
                    tourLength += after - before;
                    orOptModify(a, b, c, true);
                    return true;
                }
            }
        }
        return false;
    }

    private void orOptModify(int a, int b, int c, boolean reverse) {
        boolean drawn = false;
        if ( threadSuspended ) {
            drawn = true;
            draw(a); draw(a+1); draw(b); draw(b+1); draw(c); draw(c+1);
            if (!sleep()) { initPressed = true; return; }

            setColors(null, null, EDGE_COLOR2);
            draw(a, a+1);  draw(b, b+1);  draw(c, c+1);
            setColors(null, null, getBackground());
            if (!sleep()) { initPressed = true; return; }

            setColors(null, null, EDGE_COLOR2);
            if (!reverse) { draw(a, b+1); draw(a+1, c+1); draw(b, c); }
            else          { draw(a, b+1); draw(a+1, c); draw(b, c+1); }
            setColors(null, null, getBackground());
            if (!sleep()) { initPressed = true; return; }

            setColors(null, null, EDGE_COLOR2);
            draw(a, a+1);  draw(b, b+1);  draw(c, c+1);
            if (!reverse) { draw(a, b+1); draw(a+1, c+1); draw(b, c); }
            else          { draw(a, b+1); draw(a+1, c); draw(b, c+1); }
            setColors(null, null, getBackground());
        }
        drawTourLength("Or-opt");
        draw(a, a+1);  draw(b, b+1);  draw(c, c+1);
        if (!reverse) { draw(a, b+1); draw(a+1, c+1); draw(b, c); }
        else          { draw(a, b+1); draw(a+1, c); draw(b, c+1); }
        if ( threadSuspended || drawn ) {
            if (!sleep()) { initPressed = true; return; }
            draw(a); draw(a+1); draw(b); draw(b+1); draw(c); draw(c+1);
        }
        int len = b - a,  tmp[] = new int[len];
        for (int i = 0; i < len; i++) tmp[i] = tour[(a + 1 + i) % n];
        for (int i = b + 1; i <= c; i++) tour[(i - len) % n] = tour[i % n];
        for (int i = 0; i < len; i++) {
            tour[(c - i) % n] = (!reverse) ? tmp[i] : tmp[len - 1 - i];
        }
        if ( threadSuspended || drawn ) repaint();
    }

    private void tsp() {
        initPressed = false;

        initNoTour();
        messagePanel.removeAll();
        newLabel( " initial ", Color.blue );
        repaint();
        setSuspended( true );
        if ( !sleep() ) return;

        initTourRandom();
        messagePanel.removeAll();
        newLabel( "rand", Color.gray );
        repaint();
        setSuspended( true );
        if ( !sleep() ) return;

        newLabel( "2-opt", Color.blue );
        for ( int i=0, last=0; i<last+n; i++ ) {
            if (twoOpt(i)) {
                last = i + 1;
                if ( !sleep() ) return;
            }
            if (initPressed) return;
        }
        setSuspended( true );
        if ( !sleep() ) return;

        newLabel( "Or-opt", Color.magenta.darker() );
        for ( int i=0, last=0; i<last+n; i++ ) {
            if (orOpt(i)) {
                last = i + 1;
                if ( !sleep() ) return;
            }
            if (initPressed) return;
        }
        newLabel( "finish", Color.orange.darker() );
        System.out.println("tour length = " + tourLength + " (+" +
	       (double)((long)(tourLength-optimal)*1000/optimal)/10 + "%)" );
        //for ( int i=0; i<n; i++ ) System.out.println(tour[i]+1);
        sleepForever();
    }

    public void run() {
        super.run();

        setColors(null, null, getBackground());  // XOR 描画
        setSleep( 5000 / n );
        //setSleep( 1 );
        setReady(true);
        
        while (thread != null) {
            tsp();
            System.gc();
        }
    }
}
